#pragma once

#ifdef __GNUC__
#   pragma GCC diagnostic push
#   pragma GCC diagnostic ignored "-Wparentheses"
#endif

#include <cstdlib>
#include <ConsumerImplHelper/GenTLCamera.h>

namespace GenTLConsumerImplHelper
{

    template< const char ProducerName[], typename ProducerProxy >
    class CBlazeCameraT : public CGenTLCamera<ProducerName, ProducerProxy>
    {
        using Base = CGenTLCamera<ProducerName, ProducerProxy>;
    public:
#pragma pack(push, 1)
        struct Coord3D
        {
            float x;
            float y;
            float z;
        };
#pragma pack(pop)

    public:
        CBlazeCameraT()
        {
        }

    protected:

    };

    CIH_WEAK_SYMBOL(extern const char BlazeProducerName[]);

    class BlazeProducerProxy : public CProducerProxy <BlazeProducerName>
    {
        using Base = CProducerProxy<BlazeProducerName>;
    public:
        BlazeProducerProxy() = default;

    protected:
        bool Load(const char* dllname) override
        {
            bool loaded = false;
#ifdef CIH_WIN_BUILD
            std::string strDllName(dllname);
            const bool nameContainsSeparator = strDllName.find('/') != std::string::npos || strDllName.find('\\') != std::string::npos;
            if (!nameContainsSeparator)
            {
                std::string strDllPath(strDllName);
                // Assume that the producer is located in the directory indicated by the BASLER_TOF_BIN/BASLER_TOF_BIN64 environment variable
#  pragma warning(push, 1)
#  pragma warning(disable: 4996)
#  if CIH_WORDSIZE == 64
                const char* szBaslerTofBinDir = std::getenv("BASLER_BLAZE_BIN64");
#  else
                const char* szBaslerTofBinDir = std::getenv("BASLER_BLAZE_BIN");
#  endif

#  pragma warning(pop)
                if (NULL != szBaslerTofBinDir)
                {
                    std::string strBinDir(szBaslerTofBinDir);
                    if (*strBinDir.rbegin() != '\\' && *strBinDir.rbegin() != '/')
                    {
                        strBinDir += "\\";
                    }
                    strDllPath = strBinDir + strDllName;
                }
                loaded = m_pProducerLib->Load(strDllPath.c_str());
            }
#endif
            if (!loaded)
            {
                loaded = Base::Load(dllname);
            }
            return loaded;
        }
    };

    class CBlazeCamera : public CBlazeCameraT<BlazeProducerName, BlazeProducerProxy>
    {
        using Base = CBlazeCameraT<BlazeProducerName, BlazeProducerProxy>;
    public:
        CBlazeCamera() = default;

        static void InitProducer()
        {
#ifdef CIH_WIN_BUILD
            _putenv("IS_BASLER_BLAZE_SDK=1");
#else
           setenv("IS_BASLER_BLAZE_SDK", "1", 1);
#endif
            Base::InitProducer();
        }

        static void TerminateProducer()
        {
#ifdef CIH_WIN_BUILD
            _putenv("IS_BASLER_BLAZE_SDK=");
#else
            unsetenv("IS_BASLER_BLAZE_SDK");
#endif
            Base::TerminateProducer();
        }

        inline void Open(CameraInfo cameraInfo) override
        {
            Base::Open(cameraInfo);

            try
            {
                GenApi::CEnumerationPtr ptrGenDCStreamingMode = this->GetParameter("GenDCStreamingMode");
                if ( GenApi::IsWritable(ptrGenDCStreamingMode))
                {
                    ptrGenDCStreamingMode->FromString("Off");
                    ptrGenDCStreamingMode->GetNode()->ImposeAccessMode(GenApi::RO);
                }
            }
            catch (...)
            {
            }
        }

        inline void Open(CameraInfoKey key, const std::string& value) override
        {
            Base::Open(key, value);
        }

    };
}

#ifndef USE_DEBUG_CONSUMER
const char GenTLConsumerImplHelper::BlazeProducerName[] = "ProducerBlaze.cti";
#else
const char GenTLConsumerImplHelper::BlazeProducerName[] = "ProducerBlaze_d.cti";
#endif

template<>
#if defined (_MSC_VER)
__declspec(selectany)
#endif
std::string GenTLConsumerImplHelper::CGenTLCamera<GenTLConsumerImplHelper::BlazeProducerName, GenTLConsumerImplHelper::BlazeProducerProxy>::s_NonCachedParameter
#if defined(__GNUC__)
__attribute((weak))
#endif
= "DeviceTemperature";

#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif
