#pragma once

#include <memory>
#include <string>
#include <GenTL/GenTL_v1_5.h>
#include "DynamicLib.h"
#include <iostream>
#include <algorithm>
#if defined (_MSC_VER) && defined (_WIN32)
// We always want to link against the release version of GenApi. 
// The following define must be set before including the GenApi.h header file.
#  ifdef GENICAM_USER_ALWAYS_LINK_RELEASE
#    undef GENICAM_USER_ALWAYS_LINK_RELEASE
#  endif
#  define GENICAM_USER_ALWAYS_LINK_RELEASE 1
#  ifdef GENICAM_USER_ACCEPTS_ANY_COMPILER
#    undef GENICAM_USER_ACCEPTS_ANY_COMPILER
#  endif
#  define GENICAM_USER_ACCEPTS_ANY_COMPILER 1
#endif
#ifdef __GNUC__
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wsuggest-override"
#endif
#  include <GenApi/GenApi.h>
#ifdef __GNUC__
#pragma GCC diagnostic pop
#endif
#ifdef CIH_WIN_BUILD
#   include <delayimp.h>
#endif

namespace GenTLConsumerImplHelper
{

    namespace Impl
    {
        class Helper
        {
        public:
            static std::string GetGenicamGenTLPath()
            {
                const char* envVar = "GENICAM_GENTL" CIH_WORDSIZE_STR "_PATH";
#if defined (_MSC_VER) 
#   pragma warning(push, 1)
#   pragma warning( disable: 4996 )
#endif
                const char*  szGenicamGenTLPath = getenv(envVar);
#if defined (_MSC_VER) 
#   pragma warning(pop)
#endif
                return szGenicamGenTLPath == NULL ? "" : szGenicamGenTLPath;
            }

            static std::list<std::string> Tokenize(const std::string &source, const char *delimiter, bool keepEmpty = false)
            {
                std::list<std::string> results;

                size_t prev = 0;
                size_t next = 0;

                while ((next = source.find_first_of(delimiter, prev)) != std::string::npos) {
                    if (keepEmpty || (next - prev != 0)) {
                        results.push_back(source.substr(prev, next - prev));
                    }
                    prev = next + 1;
                }

                if (prev < source.size()) {
                    results.push_back(source.substr(prev));
                }

                return results;
            }
        };
    }

    template <const char* ProducerName >
    class CProducerProxy
    {
    public:
        CProducerProxy();
        ~CProducerProxy();

        void Init(const char* dllname);
        bool IsIntialized() const { return m_isInitialized; }
        void Terminate();

        // global functions
        GenTL::PGCInitLib                 GCInitLib;
        GenTL::PGCCloseLib                GCCloseLib;
        GenTL::PGCGetInfo                 GCGetInfo;
        GenTL::PGCGetLastError            GCGetLastError;
        GenTL::PGCReadPort                GCReadPort;
        GenTL::PGCWritePort               GCWritePort;
        GenTL::PGCReadPortStacked         GCReadPortStacked;      // GenTL v1.1
        GenTL::PGCWritePortStacked        GCWritePortStacked;     // GenTL v1.1
        GenTL::PGCGetPortURL              GCGetPortURL;           // legacy function
        GenTL::PGCGetPortInfo             GCGetPortInfo;
        GenTL::PGCGetNumPortURLs          GCGetNumPortURLs;       // GenTL v1.1
        GenTL::PGCGetPortURLInfo          GCGetPortURLInfo;       // GenTL v1.1
        GenTL::PGCRegisterEvent           GCRegisterEvent;
        GenTL::PGCUnregisterEvent         GCUnregisterEvent;

        // TL functions
        GenTL::PTLOpen                    TLOpen;
        GenTL::PTLClose                   TLClose;
        GenTL::PTLGetInfo                 TLGetInfo;
        GenTL::PTLGetNumInterfaces        TLGetNumInterfaces;
        GenTL::PTLGetInterfaceID          TLGetInterfaceID;
        GenTL::PTLGetInterfaceInfo        TLGetInterfaceInfo;
        GenTL::PTLOpenInterface           TLOpenInterface;
        GenTL::PTLUpdateInterfaceList     TLUpdateInterfaceList;

        // Interface functions
        GenTL::PIFClose                   IFClose;
        GenTL::PIFGetInfo                 IFGetInfo;
        GenTL::PIFGetNumDevices           IFGetNumDevices;
        GenTL::PIFGetDeviceID             IFGetDeviceID;
        GenTL::PIFUpdateDeviceList        IFUpdateDeviceList;
        GenTL::PIFGetDeviceInfo           IFGetDeviceInfo;
        GenTL::PIFOpenDevice              IFOpenDevice;
        GenTL::PIFGetParentTL             IFGetParentTL;          // GenTL v1.4

        // Device functions
        GenTL::PDevGetPort                DevGetPort;
        GenTL::PDevGetNumDataStreams      DevGetNumDataStreams;
        GenTL::PDevGetDataStreamID        DevGetDataStreamID;
        GenTL::PDevOpenDataStream         DevOpenDataStream;
        GenTL::PDevGetInfo                DevGetInfo;
        GenTL::PDevClose                  DevClose;
        GenTL::PDevGetParentIF            DevGetParentIF;         // GenTL v1.4

        // Data stream functions
        GenTL::PDSAnnounceBuffer          DSAnnounceBuffer;
        GenTL::PDSAllocAndAnnounceBuffer  DSAllocAndAnnounceBuffer;
        GenTL::PDSFlushQueue              DSFlushQueue;
        GenTL::PDSStartAcquisition        DSStartAcquisition;
        GenTL::PDSStopAcquisition         DSStopAcquisition;
        GenTL::PDSGetInfo                 DSGetInfo;
        GenTL::PDSGetBufferID             DSGetBufferID;
        GenTL::PDSClose                   DSClose;
        GenTL::PDSRevokeBuffer            DSRevokeBuffer;
        GenTL::PDSQueueBuffer             DSQueueBuffer;
        GenTL::PDSGetBufferInfo           DSGetBufferInfo;
        GenTL::PDSGetBufferChunkData      DSGetBufferChunkData;   // GenTL v1.3
        GenTL::PDSGetParentDev            DSGetParentDev;         // GenTL v1.4
        GenTL::PDSGetNumBufferParts       DSGetNumBufferParts;    // GenTL v1.5
        GenTL::PDSGetBufferPartInfo       DSGetBufferPartInfo;    // GenTL v1.5

        // Event functions
        GenTL::PEventGetData              EventGetData;
        GenTL::PEventGetDataInfo          EventGetDataInfo;
        GenTL::PEventGetInfo              EventGetInfo;
        GenTL::PEventFlush                EventFlush;
        GenTL::PEventKill                 EventKill;

    public:
        // Helper functions
        GenICam::gcstring                 CreateErrorMessage(const char* pText) const;

        template<class string>
        GenICam::gcstring                 CreateErrorMessage(const string& text) const;
        std::vector<char>                 RetrieveXmlFile(GenTL::PORT_HANDLE hPort, bool* isZipped);

    protected:

        // Helper functions
        static bool SplitLocalUrl(const std::string& strLocal, std::string& fullFilename, uint64_t& address, size_t& length);
        virtual bool Load(const char*);

    protected:
        bool                                    m_isInitialized;
        std::unique_ptr<Internal::CDynamicLib>  m_pProducerLib;
    };



    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    //
    //  Implementation of the CProducerProxy class
    //
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Initializes a function pointer. The function may be optionally.
#define INIT_OPT_FCT_PTR( _PTR ) \
    _PTR = (GenTL::P##_PTR) m_pProducerLib->GetFunctionPointer( #_PTR );

    // Initializes a function pointer. The pointer is checked for a NULL ptr.
#define INIT_FCT_PTR( _PTR ) \
    INIT_OPT_FCT_PTR( _PTR ); \
    if ( NULL == _PTR ) { throw RUNTIME_EXCEPTION("Invalid GenTL producer that doesn't export the " #_PTR " function."); }

    template <const char* ProducerName>
    CProducerProxy<ProducerName>::CProducerProxy()
        : GCInitLib(NULL)
        , GCCloseLib(NULL)
        , GCGetInfo(NULL)
        , GCGetLastError(NULL)
        , GCReadPort(NULL)
        , GCWritePort(NULL)
        , GCReadPortStacked(NULL)
        , GCWritePortStacked(NULL)
        , GCGetPortURL(NULL)
        , GCGetPortInfo(NULL)
        , GCGetNumPortURLs(NULL)
        , GCGetPortURLInfo(NULL)
        , GCRegisterEvent(NULL)
        , GCUnregisterEvent(NULL)
        , TLOpen(NULL)
        , TLClose(NULL)
        , TLGetInfo(NULL)
        , TLGetNumInterfaces(NULL)
        , TLGetInterfaceID(NULL)
        , TLGetInterfaceInfo(NULL)
        , TLOpenInterface(NULL)
        , TLUpdateInterfaceList(NULL)
        , IFClose(NULL)
        , IFGetInfo(NULL)
        , IFGetNumDevices(NULL)
        , IFGetDeviceID(NULL)
        , IFUpdateDeviceList(NULL)
        , IFGetDeviceInfo(NULL)
        , IFOpenDevice(NULL)
        , IFGetParentTL(NULL)
        , DevGetPort(NULL)
        , DevGetNumDataStreams(NULL)
        , DevGetDataStreamID(NULL)
        , DevOpenDataStream(NULL)
        , DevGetInfo(NULL)
        , DevClose(NULL)
        , DevGetParentIF(NULL)
        , DSAnnounceBuffer(NULL)
        , DSAllocAndAnnounceBuffer(NULL)
        , DSFlushQueue(NULL)
        , DSStartAcquisition(NULL)
        , DSStopAcquisition(NULL)
        , DSGetInfo(NULL)
        , DSGetBufferID(NULL)
        , DSClose(NULL)
        , DSRevokeBuffer(NULL)
        , DSQueueBuffer(NULL)
        , DSGetBufferInfo(NULL)
        , DSGetBufferChunkData(NULL)
        , DSGetParentDev(NULL)
        , DSGetNumBufferParts(NULL)
        , DSGetBufferPartInfo(NULL)
        , EventGetData(NULL)
        , EventGetDataInfo(NULL)
        , EventGetInfo(NULL)
        , EventFlush(NULL)
        , EventKill(NULL)
        , m_isInitialized(false)
        , m_pProducerLib(std::make_unique<Internal::CDynamicLib>())
    {
    }

    template <const char* ProducerName>
    CProducerProxy<ProducerName>::~CProducerProxy(void)
    {
        Terminate();
    }

    template <const char* ProducerName>
    void CProducerProxy<ProducerName>::Init(const char* dllname)
    {
        if (!Load(dllname))
        {
            std::ostringstream s;
            throw RUNTIME_EXCEPTION(static_cast<std::ostringstream&>(s << "Failed to load the producer file " << dllname << ": " << m_pProducerLib->GetLastErrorMsg()).str());
        }
        INIT_FCT_PTR(GCInitLib);
        INIT_FCT_PTR(GCCloseLib);
        INIT_FCT_PTR(GCGetInfo);
        INIT_FCT_PTR(GCGetLastError);
        INIT_FCT_PTR(GCReadPort);
        INIT_FCT_PTR(GCWritePort);
        INIT_FCT_PTR(GCReadPortStacked);          // GenTL v1.1
        INIT_FCT_PTR(GCWritePortStacked);         // GenTL v1.1
        INIT_FCT_PTR(GCGetPortURL);
        INIT_FCT_PTR(GCGetPortInfo);
        INIT_FCT_PTR(GCGetNumPortURLs);           // GenTL v1.1
        INIT_FCT_PTR(GCGetPortURLInfo);           // GenTL v1.1
        INIT_FCT_PTR(GCRegisterEvent);
        INIT_FCT_PTR(GCUnregisterEvent);

        INIT_FCT_PTR(TLOpen);
        INIT_FCT_PTR(TLClose);
        INIT_FCT_PTR(TLGetInfo);
        INIT_FCT_PTR(TLGetNumInterfaces);
        INIT_FCT_PTR(TLGetInterfaceID);
        INIT_FCT_PTR(TLGetInterfaceInfo);
        INIT_FCT_PTR(TLOpenInterface);
        INIT_FCT_PTR(TLUpdateInterfaceList);

        INIT_FCT_PTR(IFClose);
        INIT_FCT_PTR(IFGetInfo);
        INIT_FCT_PTR(IFGetNumDevices);
        INIT_FCT_PTR(IFGetDeviceID);
        INIT_FCT_PTR(IFUpdateDeviceList);
        INIT_FCT_PTR(IFGetDeviceInfo);
        INIT_FCT_PTR(IFOpenDevice);
        INIT_FCT_PTR(IFGetParentTL);              // GenTL v1.4

        INIT_FCT_PTR(DevGetPort);
        INIT_FCT_PTR(DevGetNumDataStreams);
        INIT_FCT_PTR(DevGetDataStreamID);
        INIT_FCT_PTR(DevOpenDataStream);
        INIT_FCT_PTR(DevGetInfo);
        INIT_FCT_PTR(DevClose);
        INIT_FCT_PTR(DevGetParentIF);             // GenTL v1.4

        INIT_FCT_PTR(DSAnnounceBuffer);
        INIT_FCT_PTR(DSAllocAndAnnounceBuffer);
        INIT_FCT_PTR(DSFlushQueue);
        INIT_FCT_PTR(DSStartAcquisition);
        INIT_FCT_PTR(DSStopAcquisition);
        INIT_FCT_PTR(DSGetInfo);
        INIT_FCT_PTR(DSGetBufferID);
        INIT_FCT_PTR(DSClose);
        INIT_FCT_PTR(DSRevokeBuffer);
        INIT_FCT_PTR(DSQueueBuffer);
        INIT_FCT_PTR(DSGetBufferInfo);
        INIT_FCT_PTR(DSGetBufferChunkData);       // GenTL v1.3
        INIT_FCT_PTR(DSGetParentDev);             // GenTL v1.4
        INIT_FCT_PTR(DSGetNumBufferParts);        // GenTL v1.5
        INIT_FCT_PTR(DSGetBufferPartInfo);        // GenTL v1.5

        INIT_FCT_PTR(EventGetData);
        INIT_FCT_PTR(EventGetDataInfo);
        INIT_FCT_PTR(EventGetInfo);
        INIT_FCT_PTR(EventFlush);
        INIT_FCT_PTR(EventKill);


        GenTL::GC_ERROR ret = GCInitLib();
        if (ret != GenTL::GC_ERR_SUCCESS)
        {
            std::ostringstream s;
            throw RUNTIME_EXCEPTION(static_cast<std::ostringstream&>(s << "Failed to initialize GenTL producer: " << std::hex << ret).str());
        }

#ifdef CIH_WIN_BUILD
        __HrLoadAllImportsForDll(DLL_NAME("GCBase"));
        __HrLoadAllImportsForDll(DLL_NAME("GenApi"));
#endif


        m_isInitialized = true;

    }

    template <const char* ProducerName>
    void CProducerProxy<ProducerName>::Terminate()
    {
        if (IsIntialized())
        {
            if (NULL != GCCloseLib)
            {
                GenTL::GC_ERROR res = GCCloseLib();
                if (res != GenTL::GC_ERR_SUCCESS)
                {
                    std::cerr << "Failed to call GCCloseLib():" << res << std::endl;
                }
            }
            m_isInitialized = false;
        }
    }

    template <const char* ProducerName>
    bool CProducerProxy<ProducerName>::SplitLocalUrl(const std::string& strLocal, std::string& fullFilename, uint64_t& address, size_t& length)
    {
        /*
        * Format of strLocal:
        *   "[///]filename.extension;address;length?SchemaVersion=1.0.0"
        *   Example: tlguru_system_rev1.xml;F0F00000;3BF
        */

        std::stringstream ss;

        // skip optional leading slashes
        std::string slashes("///");
        if (strLocal.find(slashes) == std::string::npos)
            ss << strLocal;
        else
            ss << strLocal.substr(slashes.length());

        // get filename
        std::getline(ss, fullFilename, ';');

        // get address;length from remainder
        uint64_t addr = 0;
        size_t len = 0;

        ss >> std::hex >> addr;

        // match semicolon
        if (ss.peek() == ';')
            ss.ignore();
        else
            return false;

        ss >> len;

        if (ss.fail())
            return false;

        address = addr;
        length = len;

        return true;
    }

    template <const char* ProducerName>
    bool CProducerProxy<ProducerName>::Load(const char* dllname)
    {
        bool loaded = false;
        std::string strDllName(dllname);
        const bool isGenICamTLPathSet = !Impl::Helper::GetGenicamGenTLPath().empty();
        const bool nameContainsSeparator = strDllName.find('/') != std::string::npos || strDllName.find('\\') != std::string::npos;
        std::string strDllPath(strDllName);
        loaded = m_pProducerLib->Load(strDllPath.c_str());
        if (!loaded && isGenICamTLPathSet && !nameContainsSeparator)
        {
            // Iterate over all entries in GENICAM_GENTL_PATH and try to load the desired module from there
#if defined CIH_WIN_BUILD
            const char* delimiter = ";";
#else
            const char* delimiter = ":";
#endif
            std::list<std::string> lstPathEntries = Impl::Helper::Tokenize(Impl::Helper::GetGenicamGenTLPath(), delimiter);
            for (std::list<std::string>::const_iterator it = lstPathEntries.begin(); it != lstPathEntries.end() && !loaded; ++it)
            {
                strDllPath = *it;
                if (*strDllPath.rbegin() != '\\' && *strDllPath.rbegin() != '/')
                    strDllPath += CIH_PATH_SEPARATOR;
                strDllPath += strDllName;
                loaded = m_pProducerLib->Load(strDllPath.c_str());
            }
        }
        return loaded;
    }

    template <const char* ProducerName>
    GenICam::gcstring CProducerProxy<ProducerName>::CreateErrorMessage(const char *pText) const
    {
        GenTL::GC_ERROR iErrorCode = GenTL::GC_ERR_NOT_INITIALIZED;
        std::vector<char> buf(1024);
        std::stringstream message;
        message << pText;

        size_t bufSize = buf.size();

        GenTL::GC_ERROR status = GCGetLastError(&iErrorCode, &buf[0], &bufSize);
        if (status == GenTL::GC_ERR_BUFFER_TOO_SMALL || bufSize > buf.size())
        {
            buf.resize(bufSize);
            status = GCGetLastError(&iErrorCode, &buf[0], &bufSize);
        }
        if (status == GenTL::GC_ERR_SUCCESS)
        {
            message << ": " << &buf[0];
        }
        else
        {
            message << " (Unknown reason).";
        }

        return message.str().c_str();
    }

    template <const char* ProducerName>
    template <class string>
    GenICam::gcstring CProducerProxy<ProducerName>::CreateErrorMessage(const string& text) const
    {
        return CreateErrorMessage(text.c_str());
    }



    template <const char* ProducerName >
    std::vector<char> GenTLConsumerImplHelper::CProducerProxy<ProducerName>::RetrieveXmlFile(GenTL::PORT_HANDLE hPort, bool* isZipped)
    {
        using namespace GenTL;

        std::vector<char> result;
        GC_ERROR status;

        // retrieve URL
        uint32_t numURLs = 0;
        status = GCGetNumPortURLs(hPort, &numURLs);
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(CreateErrorMessage("GCGetNumPortURLs failed"));
        }
        if (numURLs == 0)
        {
            throw RUNTIME_EXCEPTION(CreateErrorMessage("Port doesn't provide any provide URL for an XML file."));
        }

        std::vector<char> bufUrl(2048);
        size_t urlBufSize = bufUrl.size();
        const uint32_t idxUrl = 0;  // support only first URL here
        INFO_DATATYPE dataType = INFO_DATATYPE_UNKNOWN;

        status = GCGetPortURLInfo(hPort, idxUrl, URL_INFO_URL, &dataType, &bufUrl[0], &urlBufSize);
        if (status == GC_ERR_BUFFER_TOO_SMALL || urlBufSize > bufUrl.size())
        {
            bufUrl.resize(urlBufSize);
            status = GCGetPortURLInfo(hPort, idxUrl, URL_INFO_URL, &dataType, &bufUrl[0], &urlBufSize);
        }
        if (status != GC_ERR_SUCCESS)
        {
            throw RUNTIME_EXCEPTION(CreateErrorMessage("Failed to retrieve information about the XML File URL"));
        }

        // Read XML Address
        std::string strXMLAddress(&bufUrl[0]);
        std::string strLocation = strXMLAddress.substr(0, 6);
        std::transform(strLocation.begin(), strLocation.end(), strLocation.begin(),
            [](unsigned char c) { return (unsigned char)std::toupper(c); });

        if (strLocation == "LOCAL:")
        {
            std::string fullFilename;
            uint64_t addr = 0;
            size_t xmlSize = 0;
            if (!SplitLocalUrl(strXMLAddress.substr(6), fullFilename, addr, xmlSize))
            {
                std::ostringstream s;
                throw RUNTIME_EXCEPTION(static_cast<std::ostringstream&>(s << "Failed to get length and or address from URL " << strXMLAddress).str());
            }

            result.resize(xmlSize + 1);  // incl. zero termination

            status = GCReadPort(hPort, addr, &result[0], &xmlSize);
            if (status != GC_ERR_SUCCESS)
            {
                throw RUNTIME_EXCEPTION(CreateErrorMessage("Reading from XML file failed"));
            }

            // ensure zero termination
            result[xmlSize] = '\0';
        }
        else
        {
            std::ostringstream s;
            throw RUNTIME_EXCEPTION(static_cast<std::ostringstream&>(s << "The URL " << strXMLAddress << "is not supported.").str());
        }

        if (NULL != isZipped)
        {
            *isZipped = strXMLAddress.find(".zip") != std::string::npos;
        }


        return result;
    }
}
