# **Basler blaze Camera ROS Driver**

This readme describes how to manage blaze cameras with this ROS driver. The ROS driver includes the basler_blaze ROS node, which allows you to do the following:
- Start the acquisition and acquire data
- Retrieve the data (3D point cloud, intensity image, grayscale range map, color range map, and confidence map) through dedicated topics
- Access the camera parameters through dedicated services
- Trigger the execution of camera commands, e.g., the `TriggerSoftware` command

***

## **Compatibility**

This ROS driver has been tested and is compatible with Ubuntu 20.04 LTS and the ROS Noetic Ninjemys distribution.

## **Installation**

Copy this folder to the `src` folder of your catkin workspace.

Example:

```
mkdir -p $HOME/catkin_ws/src/blazeROS 
cp -r * $HOME/catkin_ws/src/blazeROS
```

Change to your catkin workspace folder. 

Example: 

```
cd $HOME/catkin_ws
```

Source the `devel/setup.bash` script.

Example:

```
source $HOME/catkin_ws/devel/setup.bash
```

Build the code using the `catkin_make` command.

```
catkin_make
```

***

## **Using the basler_blaze Node**

Before using the node, make sure that you source the `devel/setup.bash` script.

### **Starting the Node**

The node can be started by executing the following command line:

```
roslaunch basler_blaze basler_blaze.launch
```

The following node parameters can be specified when starting the node:
- `camera` (`string`): User-defined camera name (default: *camera*). This is required if you want to connect multiple cameras at the same time.
- `device_id` (`string`): Serial number of the camera that you want to connect (default: *#1*).\
If `device_id` = *#1* or if the serial number specified doesn't match any camera in the network, the node tries to connect to the first connected camera encountered in the network.
- `frame_id` (`string`): Frame ID used in the sensor messages (default: *camera_optical_frame*).

### **Available Topics**

The following topics are published by the basler_blaze node:

- **3D point cloud**\
`[user-defined camera name]/cloud` (`sensor_msgs/PointCloud2`, default: *camera/cloud*)
- **Intensity image**\
`[user-defined camera name]/intensity` (`sensor_msgs/Image`, default: *camera/intensity*)
- **Grayscale range map**\
`[user-defined camera name]/range_map` (`sensor_msgs/Image`, default: *camera/range_map*)
- **Color range map**\
`[user-defined camera name]/range_map_color` (`sensor_msgs/Image`, default: *camera/range_map_color*)
- **Confidence map**\
`[user-defined camera name]/confidence` (`sensor_msgs/Image`, default: *camera/confidence*)

**Examples:**
To visualize the nodes use the ROS application `rviz` or `image_view`:
```
rosrun image_view image_view image:=/camera/intensity
rosrun image_view image_view image:=/camera/range_map
rosrun image_view image_view image:=/camera/range_map_color
rosrun image_view image_view image:=/camera/confidence
```

Each topic listed above is linked to another topic publishing the camera info (calibration values). Each camera info topic is only published if there is at least one subscriber to the corresponding camera-data-related topic:

- **3D point cloud camera info**\
`[user-defined camera name]/cloud/camera_info` (`sensor_msgs/CameraInfo`, default: *camera/cloud/camera_info*)
- **Intensity image camera info**\
`[user-defined camera name]/intensity/camera_info` (`sensor_msgs/CameraInfo`, default: *camera/intensity/camera_info*)
- **Grayscale range map camera info**\
`[user-defined camera name]/range_map/camera_info` (`sensor_msgs/CameraInfo`, default: *camera/range_map/camera_info*)
- **Color range map camera info**\
`[user-defined camera name]/range_map_color/camera_info` (`sensor_msgs/CameraInfo`, default: *camera/range_map_color/camera_info*)
- **Confidence map camera info**\
`[user-defined camera name]/confidence/camera_info` (`sensor_msgs/CameraInfo`, default: *camera/confidence/camera_info*)

### **Available Services**

The following services are implemented by the basler_blaze node:

#### **Setting Camera Parameters**

- **Service name:** `[user-defined camera name]/set_blaze_parameter` (default: *camera/set_blaze_parameter*)

- **Request:**
  - `ParameterID` (`string`): Name of the parameter to be accessed (refer to the blaze documentation to get the names of the camera parameters).
  - `ParameterValue` (`string`): New value of the parameter specified by `ParameterID` (refer to the blaze documentation to get the possible values of the specified camera parameter). If the parameter value is numerical or boolean, you need to use the following syntax: `'!!str [value]'`

- **Response:**
  - `StatusValue` (`int`): 0 (service execution failed) or 1 (service execution succeeded).
  - `StatusMsg` (`string`): Message explaining the overall result specified by `StatusValue`.

**Examples:**
```
rosservice call /camera/set_blaze_parameter OperatingMode LongRange
rosservice call /camera/set_blaze_parameter ExposureTime '!!str 500'
rosservice call /camera/set_blaze_parameter FastMode '!!str true'
```

#### **Getting Camera Parameters**

- **Service name:** `[user-defined camera name]/get_blaze_parameter` (default: *camera/get_blaze_parameter*)

- **Request:**
  - `ParameterID` (`string`): Name of the parameter to be accessed (refer to the blaze documentation to get the names of the camera parameters).

- **Response:**
  - `ParameterValue` (`string`): Value of the parameter specified by `ParameterID`.
  - `StatusValue` (`int`): 0 (service execution failed) or 1 (service execution succeeded).
  - `StatusMsg` (`string`): Message explaining the overall result specified by `StatusValue`.

**Examples:**
```
rosservice call /camera/get_blaze_parameter ExposureTime
rosservice call /camera/get_blaze_parameter OperatingMode
```

#### **Setting Camera Data Stream Parameters**

- **Service name:** `[user-defined camera name]/set_blaze_data_stream_parameter` (default: *camera/set_blaze_data_stream_parameter*)

- **Request:**
  - `ParameterID` (`string`): Name of the data stream parameter to be accessed (refer to the blaze documentation to get the names of the camera data stream parameters).
  - `ParameterValue` (`string`): New value of the data stream parameter specified by `ParameterID` (refer to the blaze documentation to get the possible values of the specified camera data stream parameter). If the parameter value is numerical or boolean, you need to use the following syntax: `'!!str [value]'`.

- **Response:**
  - `StatusValue` (`int`): 0 (service execution failed) or 1 (service execution succeeded).
  - `StatusMsg` (`string`): Message explaining the overall result specified by `StatusValue`.

#### **Getting Camera Data Stream Parameters**

- **Service name:** `[user-defined camera name]/get_blaze_data_stream_parameter` (default: *camera/get_blaze_data_stream_parameter*)

- **Request:**
  - `ParameterID` (`string`): Name of the data stream parameter to be accessed (refer to the blaze documentation to get the names of the camera data stream parameters).

- **Response:**
  - `ParameterValue` (`string`): Value of the data stream parameter specified by `ParameterID`.
  - `StatusValue` (`int`): 0 (service execution failed) or 1 (service execution succeeded).
  - `StatusMsg` (`string`): Message explaining the overall result specified by `StatusValue`.

**Examples:**
```
rosservice call /camera/get_blaze_data_stream_parameter Statistic_Resend_Request_Count
resservice call /camera/get_blaze_data_stream_parameter Statistic_Failed_Buffer_Count
```

#### **Executing Commands**

- **Service name:** `[user-defined camera name]/execute_command` (default: *camera/execute_command*)

- **Request:**
  - `Command` (`string`): Name of the command to be executed (refer to the blaze documentation to get the names of the camera commands).

- **Response:**
  - `StatusValue` (`int`): 0 (service execution failed) or 1 (service execution succeeded).
  - `StatusMsg` (`string`): Message explaining the overall result specified by `StatusValue`.

**Examples:**
```
rosservice call /camera/execute_command TriggerSoftware
rosservice call /camera/execute_command ClearLastError
```

***

## **Node Test**
Basic unit tests have been implemented and can be executed with the following command line (it should be executed at the catkin workspace level):

```
catkin_make run_tests
```
